@@ -43,6 +43,7 @@ class Model(object):
         self.stop_training = False
 
         # Backend-dependent attributes
+        self.opt = None
         # Tensor or callable
         self.outputs = None
         self.outputs_losses = None
@@ -66,7 +67,8 @@ class Model(object):
 
         Args:
             optimizer: String. Name of optimizer.
-            lr: A Tensor or a floating point value. The learning rate.
+            lr: A Tensor or a floating point value. The learning rate. For L-BFGS, use
+                `dde.optimizers.set_LBFGS_options` to set the hyperparameters.
             loss: If the same loss is used for all errors, then `loss` is a String (name
                 of objective function) or objective function. If different errors use
                 different losses, then `loss` is a list whose size is equal to the
@@ -157,10 +159,8 @@ class Model(object):
         # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
         @tf.function
         def outputs_losses(training, inputs, targets, auxiliary_vars):
-            # TODO: Add training
-            # self.net.training = training
+            self.net.training = training
             self.net.inputs = inputs
-            self.net.targets = targets
             self.net.auxiliary_vars = auxiliary_vars
             # Don't call outputs() decorated by @tf.function above, otherwise the
             # gradient of outputs wrt inputs will be lost here.
@@ -173,11 +173,10 @@ class Model(object):
             if self.net.regularizer is not None:
                 losses += [tf.math.reduce_sum(self.net.losses)]
             losses = tf.convert_to_tensor(losses)
-            # TODO: Weighted losses
+            # Weighted losses
             if loss_weights is not None:
-                raise NotImplementedError(
-                    "Backend tensorflow doesn't support loss_weights"
-                )
+                losses *= loss_weights
+                self.losshistory.set_loss_weights(loss_weights)
             return outputs_, losses
 
         opt = optimizers.get(self.opt_name, learning_rate=lr, decay=decay)
@@ -194,7 +193,9 @@ class Model(object):
             grads = tape.gradient(total_loss, trainable_variables)
             opt.apply_gradients(zip(grads, trainable_variables))
 
-        def train_step_tfp(inputs, targets, auxiliary_vars):
+        def train_step_tfp(
+            inputs, targets, auxiliary_vars, previous_optimizer_results=None
+        ):
             def build_loss():
                 losses = outputs_losses(True, inputs, targets, auxiliary_vars)[1]
                 return tf.math.reduce_sum(losses)
@@ -202,7 +203,7 @@ class Model(object):
             trainable_variables = (
                 self.net.trainable_variables + self.external_trainable_variables
             )
-            return opt(trainable_variables, build_loss)
+            return opt(trainable_variables, build_loss, previous_optimizer_results)
 
         # Callables
         self.outputs = outputs
@@ -216,16 +217,16 @@ class Model(object):
     def _compile_pytorch(self, lr, loss_fn, decay, loss_weights):
         """pytorch"""
 
-        def outputs(inputs):
-            # TODO: Add training
-            # TODO: Use torch.no_grad() if training is False
-            inputs = torch.as_tensor(inputs)
-            inputs.requires_grad_()
-            self.net.inputs = inputs
-            return self.net(inputs)
+        def outputs(training, inputs):
+            self.net.train(mode=training)
+            with torch.no_grad():
+                return self.net(torch.as_tensor(inputs))
 
-        def outputs_losses(inputs, targets):
-            outputs_ = outputs(inputs)
+        def outputs_losses(training, inputs, targets):
+            self.net.train(mode=training)
+            self.net.inputs = torch.as_tensor(inputs)
+            self.net.inputs.requires_grad_()
+            outputs_ = self.net(self.net.inputs)
             # Data losses
             if targets is not None:
                 targets = torch.as_tensor(targets)
@@ -233,12 +234,11 @@ class Model(object):
             if not isinstance(losses, list):
                 losses = [losses]
             # TODO: regularization
-            # TODO: Weighted losses
-            if loss_weights is not None:
-                raise NotImplementedError(
-                    "Backend pytorch doesn't support loss_weights"
-                )
             losses = torch.stack(losses)
+            # Weighted losses
+            if loss_weights is not None:
+                losses *= torch.as_tensor(loss_weights)
+                self.losshistory.set_loss_weights(loss_weights)
             # Clear cached Jacobians and Hessians.
             grad.clear()
             return outputs_, losses
@@ -249,19 +249,19 @@ class Model(object):
         trainable_variables = (
             list(self.net.parameters()) + self.external_trainable_variables
         )
-        opt = optimizers.get(
+        self.opt = optimizers.get(
             trainable_variables, self.opt_name, learning_rate=lr, decay=decay
         )
 
         def train_step(inputs, targets):
             def closure():
-                losses = outputs_losses(inputs, targets)[1]
+                losses = outputs_losses(True, inputs, targets)[1]
                 total_loss = torch.sum(losses)
-                opt.zero_grad()
+                self.opt.zero_grad()
                 total_loss.backward()
                 return total_loss
 
-            opt.step(closure)
+            self.opt.step(closure)
 
         # Callables
         self.outputs = outputs
@@ -272,12 +272,21 @@ class Model(object):
         if backend_name == "tensorflow.compat.v1":
             feed_dict = self.net.feed_dict(training, inputs)
             return self.sess.run(self.outputs, feed_dict=feed_dict)
+        # tensorflow and pytorch
+        outs = self.outputs(training, inputs)
+        return utils.to_numpy(outs)
+
+    def _outputs_losses(self, training, inputs, targets, auxiliary_vars):
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(training, inputs, targets, auxiliary_vars)
+            return self.sess.run(self.outputs_losses, feed_dict=feed_dict)
         if backend_name == "tensorflow":
-            outs = self.outputs(training, inputs)
+            outs = self.outputs_losses(training, inputs, targets, auxiliary_vars)
         elif backend_name == "pytorch":
-            # TODO: training
-            with torch.no_grad():
-                outs = self.outputs(inputs)
+            # TODO: auxiliary_vars
+            self.net.requires_grad_(requires_grad=False)
+            outs = self.outputs_losses(training, inputs, targets)
+            self.net.requires_grad_()
         return utils.to_numpy(outs)
 
     def _train_step(self, inputs, targets, auxiliary_vars):
@@ -290,19 +299,6 @@ class Model(object):
             # TODO: auxiliary_vars
             self.train_step(inputs, targets)
 
-    def _run(self, fetches, training, inputs, targets, auxiliary_vars):
-        """Runs one "step" of computation of tensors or callables in `fetches`."""
-        if backend_name == "tensorflow.compat.v1":
-            feed_dict = self.net.feed_dict(training, inputs, targets, auxiliary_vars)
-            return self.sess.run(fetches, feed_dict=feed_dict)
-        if backend_name == "tensorflow":
-            outs = fetches(training, inputs, targets, auxiliary_vars)
-        elif backend_name == "pytorch":
-            # TODO: Use torch.no_grad() in _test() and predict()
-            # TODO: training, auxiliary_vars
-            outs = fetches(inputs, targets)
-        return None if outs is None else utils.to_numpy(outs)
-
     @utils.timing
     def train(
         self,
@@ -361,6 +357,8 @@ class Model(object):
                 self._train_tensorflow_compat_v1_scipy(display_every)
             elif backend_name == "tensorflow":
                 self._train_tensorflow_tfp()
+            elif backend_name == "pytorch":
+                self._train_pytorch_lbfgs()
         else:
             if epochs is None:
                 raise ValueError("No epochs for {}.".format(self.opt_name))
@@ -427,27 +425,70 @@ class Model(object):
         self._test()
 
     def _train_tensorflow_tfp(self):
-        self.train_state.set_data_train(*self.data.train_next_batch(self.batch_size))
-        results = self.train_step(
-            self.train_state.X_train,
-            self.train_state.y_train,
-            self.train_state.train_aux_vars,
-        )
-        # TODO: Loss history
-        self.train_state.epoch += results.num_iterations.numpy()
-        self.train_state.step += results.num_iterations.numpy()
-        self._test()
+        # There is only one optimization step. If using multiple steps with/without
+        # previous_optimizer_results, L-BFGS failed to reach a small error. The reason
+        # could be that tfp.optimizer.lbfgs_minimize will start from scratch for each
+        # call.
+        n_iter = 0
+        while n_iter < optimizers.LBFGS_options["maxiter"]:
+            self.train_state.set_data_train(
+                *self.data.train_next_batch(self.batch_size)
+            )
+            results = self.train_step(
+                self.train_state.X_train,
+                self.train_state.y_train,
+                self.train_state.train_aux_vars,
+            )
+            n_iter += results.num_iterations.numpy()
+            self.train_state.epoch += results.num_iterations.numpy()
+            self.train_state.step += results.num_iterations.numpy()
+            self._test()
+
+            if results.converged or results.failed:
+                break
+
+    def _train_pytorch_lbfgs(self):
+        prev_n_iter = 0
+        while prev_n_iter < optimizers.LBFGS_options["maxiter"]:
+            self.callbacks.on_epoch_begin()
+            self.callbacks.on_batch_begin()
+
+            self.train_state.set_data_train(
+                *self.data.train_next_batch(self.batch_size)
+            )
+            self._train_step(
+                self.train_state.X_train,
+                self.train_state.y_train,
+                self.train_state.train_aux_vars,
+            )
+
+            n_iter = self.opt.state_dict()["state"][0]["n_iter"]
+            if prev_n_iter == n_iter:
+                # Converged
+                break
+
+            self.train_state.epoch += n_iter - prev_n_iter
+            self.train_state.step += n_iter - prev_n_iter
+            prev_n_iter = n_iter
+            self._test()
+
+            self.callbacks.on_batch_end()
+            self.callbacks.on_epoch_end()
+
+            if self.stop_training:
+                break
 
     def _test(self):
-        self.train_state.y_pred_train, self.train_state.loss_train = self._run(
-            self.outputs_losses,
+        (
+            self.train_state.y_pred_train,
+            self.train_state.loss_train,
+        ) = self._outputs_losses(
             True,
             self.train_state.X_train,
             self.train_state.y_train,
             self.train_state.train_aux_vars,
         )
-        self.train_state.y_pred_test, self.train_state.loss_test = self._run(
-            self.outputs_losses,
+        self.train_state.y_pred_test, self.train_state.loss_test = self._outputs_losses(
             False,
             self.train_state.X_test,
             self.train_state.y_test,
@@ -490,7 +531,7 @@ class Model(object):
                     op = operator(self.net.inputs, self.net.outputs)
                 elif utils.get_num_args(operator) == 3:
                     op = operator(self.net.inputs, self.net.outputs, x)
-                y = self._run(op, False, x, None, None)
+                y = self.sess.run(op, feed_dict=self.net.feed_dict(False, x))
             elif backend_name == "tensorflow":
                 if utils.get_num_args(operator) == 2:
 
